// (C) Copyright Adolfo Di Mare 2011
// Use, modification and distribution are subject to the
// Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

// Revision history:
// Oct 2011      Adolfo Di Mare ==> Initial (non boost) version
// Jun+Ago 2012  Adolfo Di Mare ==> fractpart correction

// mnyfmt.c (C) 2011 adolfo@di-mare.com

// Define this macro if your compiler does not have a (long long) data type
// #define MNYFMT_NO_LONG_LONG

#include "mnyfmt.h"

#ifdef __cplusplus
// To compile the C source with a C++ compiler, override the file's extension.
// The GNU gcc compiler does this with option -x: gcc -x c++
#endif

/** \file   mnyfmt.c
    \brief  Implementation for \c mnyfmt().

    \author Adolfo Di Mare <adolfo@di-mare.com>
    \date   2011
*/

// Examples and unit test cases are in in file mnyfmtts.c

/** Formats and stores in \c fmtstr the money amount.
    The formatting pattern is stored in result string \c fmtstr.
    As rounding a \c (double) into an integer is almost always challenging,
    the parameters for this function are the integer and fractional parts
    of the money amount that the client programmer previously calculated.
    - Overwrites \c fmtstr with the formatted value.
    - The integer part of the value to format is \c intpart.
    - The fractional part of the currency amount to format is \c fractpart.
    - The first occurrence of the character \c dec is de decimal fraction
      separator.
    - When the decimal fraction separator character \c dec does not appear
      in \c fmtstr it is assumed to be \c '\0' (end of string character).
    - After the \c dec separator all the leading consecutive \c '9' format
      characters are substituted with the corresponding digit from \c fractpart,
      using digit zero \c '0' to fill leading format characters if needed.
    - All digits that inmediatly follow the decimal fraction separator are
      changed either to zero or to the corresponding digit taken from
      \c fractpart.
    - Both the integer part and the fractional part are filled up from the
      least significant digit to the most significant digit. This means that
      a format string like \c "9999.9999" wild yield \c "0123.0087" as result
      if \c intpart==123 and \c fractpart==87.
    - Characters trailing after the \c dec separator that are not the \c '9'
      format digit are left untouched.
    - All format characters \c '9' appearing before the decimal separator \c
      dec will be replaced by digit zero \c '0' if the corresponding digit is
      not significant.
    - When \c intpart is negative, the \c '-' sign will be place over the
      \c '9' immediately after the more significant digit.
    - Non format characters in \c fmtstr are left untouched.
    - The negative sign always is \c '-' and it is always placed on top of
      the corresponding format character.
    - Returns a pointer to the first significant digit in the formatted string,
      or the \c '-' sign if the formatted value is negative.
    - Returns \c (char*)(0) when the formatted value does not fit within
      \c strlen(fmtstr) characters.
    - Before storing the format string in \c fmtstr, the programmer must ensure
      that \c fmtstr is big enough to hold the format string.

    \remark
    - This routine basically substitutes each \c '9' character in \c fmtstr for
      its corresponding decimal digit, or \c '0' when it is not a significant
      digit. All other characters remain untouched.
    - As it happens with many C functions, before invocation the programmer
      must be sure that \c fmtstr is big enough to copy on it the complete
      format string, as otherwise memory beyond \c fmtstr would be overwritten.
    - There is no \c (wchart_t) version for this function, as it is meant to
      place digits in a formatting string. After placement, the result string
      can be converted to other forms.

    \dontinclude mnyfmtts.c
    \skipline    test.example
    \until       }}

    \see mnyfmtts.c
*/
char* mnyfmt(char *fmtstr, char dec, mnyfmt_long intpart, unsigned fractpart) {
    #ifndef __cplusplus
        const int false = 0; const int true = !false;
    #endif
    char *pCut, *pSign, *p;
    union { unsigned Int; unsigned Frac; } n; // unnecesary but cute
    unsigned i, nNines;
    int  isPositive; // (0<=intpart)
    char digit[ 8 * (  sizeof(mnyfmt_long) > sizeof(fractpart)
                       ? sizeof(mnyfmt_long)
                       : sizeof(fractpart)
                    ) / 3 ];
    // ensure that digit[] can hold more than 2 digits
    typedef int check_digit_size[ ( (2<=sizeof(fractpart)) ? 1 : -1 ) ];

    if (!( fmtstr) ) { return 0; } // null pointer mistake
//  if (!(*fmtstr) ) { return 0; } // null string mistake
    if ( 0 <= intpart ) { isPositive = true;}
    else {
        isPositive = false;
        intpart = -intpart;
        if ( intpart<0 ) { return 0; } // -LONG_LONG_MAX-1 case
    }

    // determine dec position
    pCut = fmtstr; // points to the decimal separator (or eos)
    nNines = 0; // # of mnyfmt_format_chars in fmtstr before decimal separator
    while ( *pCut!=0 ) {
        if ( *pCut==mnyfmt_format_char ) { ++nNines; } // count
        if ( *pCut==dec ) { break; }    // mark
        ++pCut;
    }
    if ( pCut==fmtstr || nNines == 0 ) { return 0; } // NULL error

    // make sure that fractpart fits within the format string
    p = ( (*pCut==0) ? pCut : pCut+1 );
    n.Frac = fractpart;
    while (*p==mnyfmt_format_char) {
        n.Frac /= 10;
        ++p;
    }
    if ( n.Frac!=0 ) { return 0; }; // not enough digits in format string

    // store intpart's digit in array digit[]
    if ( intpart==0 ) {
        n.Int = 1; // number of digits in intpart
        digit[0] = 0;
    }
    else {
        n.Int = 0;
        do {
            digit[n.Int] = intpart % 10;
            intpart /= 10;
            ++n.Int;
        } while ( intpart!=0 );
    }

    // check that fmtstr has enough mnyfmt_format_chars
    if ( isPositive ) {
        if ( nNines < n.Int ) { return 0; }
    }
    else { // use one mnyfmt_format_char for '-' sign
        if ( nNines < n.Int+1 ) { return 0; }
    }

    // store digit[] digits into fmtstr[]
    p = pCut;
    pSign = 0; // pSign && pSignificative
    for ( i=0; i<nNines; ++i ) {
        while ( *p!=mnyfmt_format_char ) { --p; } // backtrack to previous one
        if ( i<n.Int ) {
            *p = digit[i]+'0';
            pSign = p; // points to the most significative digit
        }
        else if ( ! isPositive ) { // isNegative ==> store '-'
            *p = '-';
            pSign = p;
            isPositive = true;
        }
        else {
            *p = '0'; // replace leading mnyfmt_format_chars with '0'
        }
        --p;
    }

    // deal with the fractpart
    if ( *pCut==0 ) {
        // eos ==> ignore fractpart
    }
    else {
        p = pCut+1;
        if ( (*p) != mnyfmt_format_char ) {
            // no format for fraction ==> ignore fractpart
        }
        else {
            // find last mnyfmt_format_char
            for ( ; (*p)==mnyfmt_format_char; ++p ) {}
            p--;

            // store fractional part backwards
            while ( p!= pCut ) {
                *p = (fractpart % 10) + '0';
                --p;
                fractpart /= 10; // yields 0 if there are many fractions
            }
        }
    }

    return pSign;

#if 0
/*  When formatting (455,12) using many many digits for fractpart
    the result returned is wrong, because a lot of leading zeroes
    should be provided.
    - Trailing digits in \c fractpart are truncated when not enough \c '9'
      format characters follow the \c dec decimal fraction separator.

    The correct specificaction should be this:
    - All digits that inmediatly follow the decimal fraction separator are
      changed either to zero or to the corresponding digit taken from
      \c fractpart.
*/
    if (false) { // broken example
        strcpy(            fmtstr ,  "999.999999999" );
        sgn = mnyfmt( fmtstr ,  '.', 455,12 )
        assertTrue( eqstr( fmtstr ,  "455.000000012" ) );
    }

    /*  OLD/WRONG implementation */
    // deal with the fractpart
    n.Frac = 0; // number of digits in fractpart
    if ( *pCut==0 || fractpart==0 ) {
        // eos || no fraction ==> ignore fractpart
    }
    else if ( *(pCut+1) != mnyfmt_format_char ) {
        // no format for fraction ==> ignore fractpart
    }
    else {
        do { // caveat: these digits come up backwards
            digit[n.Frac] = fractpart % 10;
            fractpart /= 10;
            ++n.Frac;
        } while ( fractpart!=0 );
    }
    if ( *pCut==0 ) { return pSign; }

    // store fractional digits from digits[] into fmtstr
    p = pCut + 1; // first mnyfmt_format_char
    i = n.Frac;
    while ( i>0 && *p==mnyfmt_format_char ) {
        --i;
        *p = digit[i]+'0';
        ++p;
    }
    // fill up trailing mnyfmt_format_chars with zeroes
    while ( *p==mnyfmt_format_char ) { *p = '0'; ++p; }

    return pSign;
#endif
}

// EOF: mnyfmt.c
