// (C) Copyright Adolfo Di Mare 2011
// Use, modification and distribution are subject to the
// Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

// Revision history:
// Oct 2011      Adolfo Di Mare ==> Initial (non boost) version
// Jun+Ago 2012  Adolfo Di Mare ==> fractpart correction

// mnyfmtdd.c (C) 2011 adolfo@di-mare.com

// Define this macro if your compiler does not have a (long long) data type
// #define MNYFMT_NO_LONG_LONG

#include "mnyfmt.h"

/** \file   mnyfmtdd.c
    \brief  Implementation for \c mnyfmtdd().

    \author Adolfo Di Mare <adolfo@di-mare.com>
    \date   2011
*/

#ifdef __cplusplus
// To compile the C source with a C++ compiler, override the file's extension.
// The GNU gcc compiler does this with option -x: gcc -x c++
#endif

// Examples and unit test cases are in in file mnyfmtts.c

/** Wrapper around \c mnyfmt() to use a \c modf() separated double value.
    - Oftentimes, parameters \c intdouble and \c fractdouble are result of \c modf().
    - Truncates the number of decimals in \c fractdouble to the number of
      decimlas used in the format string \c fmtstr.
    - Ignores the sign in \c fractdouble.
    - Never invokes \c modf().

    \dontinclude mnyfmtts.c
    \skipline    test.mnyfmtdd()
    \until       }}
    \dontinclude mnyfmtts.c
    \skipline    test.modf()
    \until       }}
*/
char* mnyfmtdd(char *fmtstr, char dec, double intdouble, double fractdouble) {
    const char *m;
    mnyfmt_long intpart = intdouble;

    if ( fractdouble<0.0 ) {
        fractdouble = -fractdouble;
    }
    // skip to decimal separator
    for ( m=fmtstr; (*m)!=dec; ++m) { }

    if ( *m==0 ) { // no fractional formatting
        return mnyfmt( fmtstr, dec, intpart, 0 );
    }
    else {
        // count the number of fractional digits in fmtstr
        ++m;
        unsigned pow10 = 1;
        while ( (*m)==mnyfmt_format_char ) {
            pow10 *= 10;
            ++m;
        }

        return mnyfmt( fmtstr, dec, intpart, (unsigned)(fractdouble*pow10) );
    }
}


/*  THESE ARE NOT AVAILABLE FOR OLD C COMPILERS
    ===========================================

    Function: int fesetround (int round)
    http://www.gnu.org/software/libc/manual/html_node/Rounding.html

    Changes the currently selected rounding mode to round. If round does not
    correspond to one of the supported rounding modes nothing is changed.
    fesetround returns zero if it changed the rounding mode, a nonzero value
    if the mode is not supported.

    You should avoid changing the rounding mode if possible. It can be an
    expensive operation; also, some hardware requires you to compile your
    program differently for it to work. The resulting code may run slower.
    See your compiler documentation for details.

    http://www.gnu.org/software/libc/manual/html_node/Floating-Point-Parameters.html
    These macro definitions can be accessed by including the header file
    float.h in your program.

    FLT_ROUNDS This value characterizes the rounding mode for floating point addition.

    Function: long double rintl (long double x)
    http://www.gnu.org/software/libc/manual/html_node/Rounding-Functions.html

    These functions round x to an integer value according to the current
    rounding mode. See Floating Point Parameters, for information about the
    various rounding modes. The default rounding mode is to round to the
    nearest integer; some machines support other modes, but round-to-nearest
    is always used unless you explicitly select another.

    If x was not initially an integer, these functions raise the inexact
    exception.
*/

// EOF: mnyfmtdd.c
