// (C) Copyright Adolfo Di Mare 2011
// Use, modification and distribution are subject to the
// Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

// str2mnyCE.c (C) 2013 adolfo@di-mare.com

// Define this macro if your compiler does not have a (long long) data type
// #define MNYFMT_NO_LONG_LONG

#include "mnyfmt.h"

#ifdef __cplusplus
// To compile the C source with a C++ compiler, override the file's extension.
// The GNU gcc compiler does this with option -x: gcc -x c++
#endif

/** \file   str2mnyCE.c
    \brief  Implementation for \c str2mnyCE().

    \author Adolfo Di Mare <adolfo@di-mare.com>
    \date   2011
*/

/** Returns an integer value that corresponds to 'amount', scaled 10^CE digits.
    - For "  -102,455.87" returns '-102_455_8700' when CE==4 && dec=='.'.
    - For "  -102,455.07" returns '-102_455_07' when CE==2 && dec=='.'.
    - For "  -102,455.87" returns '-102_455' when CE==0 && dec=='.'.
    - The decimal separator is 'dec'.
    - CE is the 'currency exponent'.
    - Requires 'CE<=6' <--> scale up to 10^6 <--> 6 decimal digits.

    \dontinclude mnyfmtts.c
    \skipline    test::str2mnyCE()
    \until       }}

    \see str2mnyCE.c
*/
mnyfmt_long str2mnyCE( const char * amount , char dec , unsigned CE ) {
    #ifndef isdigit
        #define isdigit(ch) ('0'<=(char)(ch) && (char)(ch)<='9')
        #define remove_isdigit
    #endif
    static mnyfmt_long TENPOW[6+1] = { 1,10,100,1000,10000,100000,1000000 };
    if ( CE>(-1+sizeof(TENPOW)/sizeof(*TENPOW)) ) { return 0; }

    const char *p = amount;         // "  -102,455.87"
    int isNegative = 0; // false
    while ( (*p!=0) && !isdigit(*p) && *p!=dec ) { isNegative = (*p=='-'); ++p; }
    if ( *p==0 ) { return 0;  }

    mnyfmt_long nInt = 0;
    while ( *p!=0 && *p!=dec ) {
        if ( isdigit(*p) ) { nInt = nInt*10 + (*p-'0'); }
        ++p;
    }

    mnyfmt_long nFrac = 0;
    if ( *p!=0 ) { ++p; } // skip 'dec'
    unsigned i;
    for ( i=0; i<CE; ++i ) {
        nFrac = nFrac*10;
        if ( isdigit(*p) ) {  nFrac += (*p-'0'); ++p; }
        else { /* nada -- nothing */ }
    }

    nInt = nInt * TENPOW[CE] + nFrac;
    if ( isNegative ) { nInt = -nInt; }
    return nInt;

    #ifdef remove_isdigit
        #undef isdigit
    #endif
    #undef remove_isdigit
}

// EOF: str2mnyCE.c
